from Globals import InitializeClass, DTMLFile, ImageFile
from OFS.SimpleItem import Item
from OFS.ObjectManager import ObjectManager
from OFS.PropertyManager import PropertyManager
from AccessControl.Role import RoleManager
from Products.ZCatalog.ZCatalog import ZCatalog
from Products.ZCTextIndex.Lexicon import Lexicon
from Products.ZCTextIndex.ZCTextIndex import ZCTextIndex, PLexicon, manage_addLexicon
from Products.ZCTextIndex.PipelineFactory import element_factory
from OFS.CopySupport import CopyContainer
from DateTime import DateTime
import string
import re
import calendar
import Acquisition


# List of BloggemItem subclasses (tuple pairs of friendly name and meta_types). 
# This list is used for reserved names and for building queries from url strings.
# The first item in the tuple is used in URL strings to build queries.  The first
# item is also used for the "name" attached to icons.  The second item represents
# the meta_type attached to the first item.
bloggemItemTypes = [
                    ('all', ''),
                    ('articles', 'Bloggem Article'),
                    ('books', 'Bloggem Book'),
                    ('links', 'Bloggem Link'),
                    ('email', 'Bloggem Email'),
                    ('organizations', 'Bloggem Organization'),
                    ('events', 'Bloggem Event'),
                    ('addresses', 'Bloggem Address')
                    ]                   


class Empty(Item):
    """ This is an empty class. """
    pass

class Bloggem(ObjectManager, PropertyManager, RoleManager, Item):

    """
    Bloggem is the root object of the Bloggem Product.  A Bloggem
    contains BloggemFolders, which contain BloggemItems.

    """

    meta_type = 'Bloggem'
    
    manage_options = (
        {'label': 'Contents', 'action': 'manage_main'},
        {'label': 'Properties', 'action': 'manage_editBloggemForm',},
        {'label': 'View', 'action': 'index_html',},
        ) + Item.manage_options 
    
    
    def __init__(self, id, title, description, meta_title, meta_description, meta_keywords, item_ids, REQUEST={}):
        """Inits the product with default values"""
        self.id = id
        self.title = title
        self.description = description
        self.meta_description = meta_description
        self.meta_keywords = meta_keywords
        self.meta_title = meta_title
        # Initialize list for item id's
        # item_ids is list of ids used for Bloggem Items.  This list is incremented
        # each time a Bloggem Item is added
        self.item_ids = item_ids
        # Add catalog, lexicon, indexes and metadata.
        self.addCatalog()

        
# Begin Catalog functions       
            
    # Create catalog
    #
    def addCatalog(self):
        """
        Add a default catalog with 'Catalog' id.
        """
        self._setObject('Catalog', ZCatalog('Catalog', title='Catalog', container=self))
        cat = self._getOb('Catalog')
        self.addLexicon(cat)
        self.addIndexes(cat)
        self.addMetas(cat)

        
    # Create default Lexicon
    #
    def addLexicon(self, cat):
        """
        Add a default lexicon with 'Lexicon' id.
        """
        wordSplitter = element_factory.instantiate('Word Splitter', 'HTML aware splitter')
        caseNormalizer = element_factory.instantiate('Case Normalizer', 'Case Normalizer')
        cat._setObject('Lexicon', PLexicon('Lexicon', 'Lexicon', wordSplitter, caseNormalizer))


    # Create indexes
    #
    def addIndexes(self, cat):
        """
        Add default indexes to Catalog.
        """
       
        cat.addIndex('id', 'FieldIndex')
        
        # Add ZCTextindexes
        title_extras = Empty()
        title_extras.doc_attr = 'title'
        title_extras.index_type = 'Okapi BM25 Rank'
        title_extras.lexicon_id = 'Lexicon'
        cat.addIndex('title', 'ZCTextIndex', title_extras)

        content_extras = Empty()
        content_extras.doc_attr = 'content'
        content_extras.index_type = 'Okapi BM25 Rank'
        content_extras.lexicon_id = 'Lexicon'
        cat.addIndex('content', 'ZCTextIndex', content_extras)      

        title_and_content_extras = Empty()
        title_and_content_extras.doc_attr = 'titleAndContent'
        title_and_content_extras.index_type = 'Okapi BM25 Rank'
        title_and_content_extras.lexicon_id = 'Lexicon'
        cat.addIndex('titleAndContent', 'ZCTextIndex', title_and_content_extras)         
        
        # Add other indexes
        cat.addIndex('create_date', 'DateIndex')
        cat.addIndex('path', 'PathIndex')
        cat.addIndex('meta_type', 'TextIndex')
        cat.addIndex('sticky', 'FieldIndex')
        cat.addIndex('add_state', 'FieldIndex')
        cat.addIndex('add_city', 'TextIndex')
        cat.addIndex('item_date_time', 'DateIndex')
        
        
    # Add metadata items
    #
    def addMetas(self, cat):
        """
        Add default metadatas to Catalog.
        """
        cat.manage_addColumn('id')
        cat.manage_addColumn('title')
        cat.manage_addColumn('meta_type')


# End Catalog functions 

    def manage_editAction(self, title, description, meta_title, meta_description, meta_keywords, RESPONSE=None):
        """Handles editing for a Bloggem."""
        self.title = title
        self.description = description
        self.meta_description = meta_description
        self.meta_keywords = meta_keywords
        self.meta_title = meta_title
        self._p_changed = 1
        RESPONSE.redirect('manage_editBloggemForm')
        

    def all_meta_types(self):
        """
        Restricts object types that can be added to a Bloggem.
        Only BloggemFolders can be added to a Bloggem object.
        """
        allowedMetaTypes = ('Bloggem Category',)
        result = []
        import Products
        for metaType in Products.meta_types:
            if metaType['name'] in allowedMetaTypes:
                    result.append(metaType)
        return result
    

    def __before_publishing_traverse__(self, object, REQUEST):
        """
        Strips query parameters from url string and performs additional
        pre-publishing processing, such as creating a list of breadcrumbs.
        """
        
        # Check for 'already_traversed' flag in REQUEST.
        # This flag is set to 'true' after a successful search
        # for params in the URL string.  Subsequent traversal
        # for the same page skips the parameter search.

       
        if REQUEST.has_key('already_traversed') and REQUEST.get('already_traversed') == "true":
            
            # Skip parameter check -- traverse normally.
            return
        
        # This is first traversal for this document.
        # Search for params and build breadcrumbs.
        else:
            
            # Strip params from URL
            traversal_stack = REQUEST['TraversalRequestNameStack']
            urlParams = []
            if REQUEST.has_key('PATH_INFO'):
                urlRegexp = re.compile ("/([^/]+)")
                urlString = REQUEST.get('PATH_INFO')
                urlParams = urlRegexp.findall(urlString)
                
            else:
                # Traverse normally. This provides handling if somehow
                # REQUEST does not have 'PATH_INFO'.
                return 
            
            # srchParams is dictionary that is used for query
            srchParams = {}
            # breadcrumbs is list of path and title tuples
            breadcrumbs = []
            # matchedParams is list for parameters matched in URL string
            # This will be used to check for non-matched parameters
            # once the URL string has been completely parsed.
            matchedParams = []
            # matchedPath is list of folders matched in traversal string.
            # It is used to modify the resulting traversal
            matchedPath = []
            # Set Bloggem title and id in REQUEST (for checking when displaying breadcrumbs)
            REQUEST.set('bloggem_id', self.id)
            REQUEST.set('bloggem_title', self.title)
            # Set bloggemItemTypes in REQUEST for navigation by icons
            REQUEST.set('itemTypes', bloggemItemTypes)
            
            # Set default values
            # Create default for search and add to REQUEST
            srchParams = {'create_date' : {'query' : DateTime(), 'range' : 'range:max'}, 'sort_on' : 'create_date', 'sort_order' : 'descending'}
            REQUEST.set('srchParams', srchParams)
            
            # Check for virtual host monster.  
            # It seems that VHM adds '/' and '[Virtual Host Monster id]' to the
            # traversal stack only if VHM is pointed directly at a Bloggem.  Thus, 
            # if a VHM is found, and it is pointed directly at a Bloggem, these
            # VHM params must be added back to the traversal_stack if only matched
            # params are found.
            vhmParams = []
            if 'VirtualHostRoot' in urlParams:
                matchedParams.append('VirtualHostRoot')
                if len(traversal_stack) > 1 and traversal_stack[-2] == '/':
                    vhmParams.append(traversal_stack[-2])
                    vhmParams.append(traversal_stack[-1])
   
            
            # Get params from urlParams and build query.
            # Strip path element(s) preceding potential parameters. 
            rootName = self.id
            # Set basepath in REQUEST
            absBasePath = self.absolute_url
            basePath = string.split(urlString, rootName)[0] + rootName
            REQUEST.set('basePath', absBasePath)
            # Set home path in breadcrumbs
            breadcrumbs.append(('', 'Home'))
            # Set default search path in REQUEST
            REQUEST.set('query_path', '/') 
            if rootName in urlParams:            
                startPos = urlParams.index(rootName) + 1 
                # remove base from urlParams
                del urlParams[0:startPos]        
                    
            
            # Check remaining urlParams for potential parameters
            if len(urlParams) > 0:

                # Check for meta_type parameter.
                crumbType = []
                for thisType in bloggemItemTypes:
                    if thisType[0] in urlParams:
                        # Add to matchedParams
                        matchedParams.append(thisType[0])
                        # Save Breadcrumb
                        thisTypeName = string.capwords(thisType[0])
                        crumbType = [thisType[0], thisTypeName]
                        # Strip meta_type from urlParams
                        meta_typePos = urlParams.index(thisType[0])
                        del urlParams[meta_typePos]
                        # Add to srchParams
                        srchParams['meta_type'] = { 'query' : thisType[1], 'operator' : 'and'}
                        break
                   
                # Check for date parameter.
                crumbDate = []
                dateRegexp = re.compile ("^\d?\d-\d?\d-\d\d\d\d")
                for dateParam in urlParams:
                    if dateRegexp.match(dateParam):
                        if dateRegexp.match(dateParam).end() == len(dateParam):
                            # Add to matched Params
                            matchedParams.append(dateParam)
                            # Parse date for DateTime query
                            replDashes = re.compile("-")
                            srchDate = replDashes.sub("/", dateParam)
                            srchDate = DateTime(srchDate).latestTime()
                            # Create query string and add to srchParams
                            srchParams['create_date']['query'] = srchDate
                            matchedParams.append(dateParam)
                            # Save Breadcrumb
                            crumbDate = [dateParam, dateParam]
                            #Strip date from urlParams
                            datePos = urlParams.index(dateParam)
                            del urlParams[datePos]
                            # Set calendar_date in REQUEST
                            REQUEST.set('calendar_date', srchDate)
                            break
                        
                # Remaining path elements consist of categories (if any)
                # and subcategories (if any).
                # Collect path elements for query.
                categoryPath = ""
                if len(urlParams) > 0:
                    # Convert remaining urlParams to a relative path and title
                    convertedUrlParams = getFolderTitles(self, urlParams)
                    folderList = convertedUrlParams[0]
                    matchedPath = convertedUrlParams[1]
                    # Use last path in folderList as search path
                    if len(folderList) > 0:
                        categoryPath = folderList[-1][0]
                        # Append categories to breadcrumbs
                        for folder in folderList:
                            breadcrumbs.append((folder[0], folder[1]))
                        # Add to srchParams
                        srchParams['path'] = {'query':categoryPath, 'level':-1}
                        # Set folder path in REQUEST
                        REQUEST.set('query_path', categoryPath)                         
                        # Add matched folders to UrlParams
                        for matchedFolder in matchedPath:
                            matchedParams.append(matchedFolder)
                    
                # Set path query in REQUEST    
                REQUEST.set('query_path', categoryPath)
                # Set query for sticky items in REQUEST
                stickyQuery = {'path' : {'query':categoryPath, 'level':-1}, 'sticky' : 'on'}
                REQUEST.set('stickyQuery', stickyQuery)
                
        
                # Append type to breadcrumbs
                if len(crumbType) > 0:
                    crumbType[0] = breadcrumbs[-1][0] + "/" + crumbType[0]
                    breadcrumbs.append((crumbType[0], crumbType[1]))
                
                # Append date to breadcrumbs
                if len(crumbDate) > 0:
                    crumbDate[0] = breadcrumbs[-1][0] + "/" + crumbDate[0]
                    breadcrumbs.append((crumbDate[0], crumbDate[1]))
                
                # Set breadcrumbs in REQUEST.
                REQUEST.set('breadcrumbs', breadcrumbs)
                REQUEST.set('already_traversed', 'true')
                        
                # Handle traverse that has matching params.
                if len(srchParams) > 0:
                    
                    # Set search params in REQUEST.
                    REQUEST.set('srchParams', srchParams)
                    
                    # Check for non-matching params in url_params.
                    # If found, traverse normally.
                    for n in urlParams:

                        if n not in matchedParams:
                            return
                    
                    # Path consists only of matched params.  Reverse
                    # matchedPath and set as traversal stack.
                    newStack = []
                    if len(matchedPath) > 0:
                        for folder in matchedPath:
                            newStack.insert(0, folder)
                            
                    # Add VHM Params if any.        
                    newStack = newStack + vhmParams
                    pass
                            
                    REQUEST['TraversalRequestNameStack'] = newStack
                    return                
                    
            # No params.  Traverse normally.
            # Set breadcrumbs in REQUEST.
            REQUEST.set('breadcrumbs', breadcrumbs) 
            REQUEST.set('already_traversed', 'true')
            return  
        
  
    def calendarGetMonth(self, REQUEST=None):
        """ 
        Reformats calendar module output, and sets a link param for
        days that have entries.
        """
        
        # Thanks to runyaga at ZopeLABS for the calendar formatting code.
        
        if REQUEST.has_key('calendar_date'):
            calendar_date = REQUEST.get('calendar_date')
        else:
            calendar_date = DateTime()
            
        themonth = calendar_date.month()
        theyear = calendar_date.year()
        
        # Build calendar as tuples of weeks.
        calendar.setfirstweekday(6) #start day  Mon(0)-Sun(6)
        _cal = {}
        _cal['month']=calendar.month_name[themonth]
        _cal['dayname_sequence']=calendar.weekheader(1).strip().split()
        _cal['weeks_by_days']=None #sequence of weeks by sequence of days
        _weeks=()
    
        for aweek in calendar.monthcalendar(theyear, themonth):
            l_days=()
            for x in range(3, 22, 3):
                
                # Convert day number int to string.
                day_number = calendar.week(aweek, 2)[x-3:x].strip()
                match_date_param = None
                
                # Check for matching Catalog entries for this date.
                if day_number != '' :
                    date_string = `theyear` + "/" + `themonth` + "/" + day_number
                    late_date = DateTime(date_string).latestTime()
                    early_date = DateTime(date_string).earliestTime()
                    date_query = self. Catalog(create_date={'query': (early_date, late_date), 'range':'min:max'})
                    
                    # If match found, set queriable date param.
                    if len(date_query) > 0:
                        match_date_param = `themonth` + "-" + day_number + "-" + `theyear`
                        
                date_pair = ((day_number, match_date_param),)
                l_days = l_days + date_pair
                 
            _weeks=_weeks+(l_days,)
    
        _cal['weeks_by_days']=_weeks
        
        # Set REQUEST params for calendar.
        REQUEST.set('calendar_date', calendar_date)
        self.getMonthNav(calendar_date, REQUEST)
        return _cal
    
    
    def getMonthNav(self, current_date=DateTime(), REQUEST=None):
        """
        Given a DateTime, sets REQUEST values for string 
        formatted dates for the first day of the previous 
        and next months.
        """
        current_month = current_date.month()
        current_year = current_date.year()
        
        if current_month == 12:
            next_month = 1
            next_year = current_year + 1
        else:
            next_month = current_month + 1
            next_year = current_year
        if current_month == 1:
            previous_month = 12
            previous_year = current_year - 1
        else:
            previous_month = current_month -1
            previous_year = current_year
            
        previous_month = `previous_month` + "-1" + "-" + `previous_year`
        next_month = `next_month` + "-1" + "-" + `next_year`
        
        REQUEST.set('previous_month', previous_month)
        REQUEST.set('next_month', next_month)
        
        return REQUEST
    
    
    
    def shortEntry(self, content, entry_url, entry_length):
        """
        Takes content from a BloggemItem, shortens it, and adds an STX-compatible
        absolute URL.
        """
        split_content = content.split()[:entry_length]
        if len(split_content) < entry_length:
            short_content = content
        else:
            short_content = string.join(split_content) + '&hellip; ' + '"More":' + entry_url + ' ' 

        return short_content
    
    def titleAndContent(self):
        """
        titleAndContent is used to create an index containing the text of 
        both the title and content fields.  This allows a single ZCTextIndex
        search to be performed on both.
        """
        return '%s %s' % (self.title,self.content)
    
  
    def manage_renameObjects(self, ids=[], new_ids=[], REQUEST=None):
        """
        Checks for reserved names on object rename.
        """
        # This only handles renaming for BloggemFolders
        # because this method is not caught when renaming
        # Bloggems.
        
        # Check for reserved id.
        for check_id in new_ids:
            checkReservedName(self, check_id)
        
        # If name not reserved, call rename method.
        CopyContainer.manage_renameObjects(self, ids, new_ids, REQUEST=None)
        return self.manage_main(self, REQUEST, update_menu=1)
    
    
    #
    # Bloggem Management Pages
    #
    manage_editBloggemForm = DTMLFile('www/editBloggemForm', globals())
    manage_css = DTMLFile('www/manage_css', globals())
    add_edit_main = DTMLFile('www/add_edit_main', globals())  
    edit_bloggem_fields = DTMLFile('www/edit_bloggem_fields', globals()) 
    check_virtual_host = DTMLFile('www/check_virtual_host', globals())
    
    #
    # Bloggem Display Pages
    #
    index_html = DTMLFile('www/index_html', globals()) 
    disp_header = DTMLFile('www/disp_header', globals())
    disp_footer = DTMLFile('www/disp_footer', globals())
    disp_calendar = DTMLFile('www/disp_calendar', globals())
    disp_css = DTMLFile('www/disp_css', globals())
    disp_entry_long = DTMLFile('www/disp_entry_long', globals())
    disp_entry_short = DTMLFile('www/disp_entry_short', globals())
    disp_single = DTMLFile('www/disp_single', globals())
    disp_home_url = DTMLFile('www/disp_home_url', globals())
    disp_title = DTMLFile('www/disp_title', globals())
    disp_nav = DTMLFile('www/disp_nav', globals())
    disp_breadcrumbs = DTMLFile('www/disp_breadcrumbs', globals())
    disp_recurse_folder = DTMLFile('www/disp_recurse_folder', globals())
    disp_cat_hdr = DTMLFile('www/disp_cat_hdr', globals())    
    disp_date_select = DTMLFile('www/disp_date_select', globals())
    disp_state_select = DTMLFile('www/disp_state_select', globals())
    disp_search = DTMLFile('www/disp_search', globals())

    
    #
    # Bloggem Images
    #
    clear_gif = ImageFile('www/gif/clear.gif', globals())
    nav_back = ImageFile('www/gif/nav_back.gif', globals())
    nav_current = ImageFile('www/gif/nav_current.gif', globals())
    nav_forward = ImageFile('www/gif/nav_forward.gif', globals())
    
#    
# Bloggem Factory
#
addBloggemForm = DTMLFile('www/addBloggemForm', globals()) 

    
def addBloggemAction(self, title, description=None, meta_title=None, meta_description=None, meta_keywords=None, id = None, item_ids=[1], REQUEST=None):
    "Creates a Bloggem."
    checkReservedName(self, id)
    self._setObject(id, Bloggem(id, title, description, meta_title, meta_description, meta_keywords, item_ids))
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

    
def checkReservedName(self, id):
    """
    Checks (Bloggem & BloggemFolder) for a reserved title.
    Raises error if submitted title is reserved.
    """
    reservedTitles = []
    for friendlyName in bloggemItemTypes:
        reservedTitles.append(friendlyName[0])
        
    if id not in reservedTitles:
        return
    
    else:
        strTitles = repr(reservedTitles)
        strTitles = strTitles[1:-1]
        errMsg = '"' + id + '" is a reserved name.  Bloggems and Categories may not be named: ' + strTitles + '.'
        raise errMsg, id   
    

        
def getFolderTitles(self, folder_list):
    """
    Function that takes a hierachical list of folders
    (Bloggem Categories) and returns a list of tuples consisting
    of the relative path for the folder and the title_or_id objects.
    Also returned is a list of matched folder names.
    """

    foldersWithTitles = []
    matchedFolders = []
    myPath = ""
    context = self
    for folderId in folder_list:
        containedCategories = context.objectItems(['Bloggem Category'])
        for category in containedCategories:
            if folderId == category[0]:
                myPath = myPath + "/" + folderId
                myTitle = category[1].title_or_id
                foldersWithTitles.append((myPath, myTitle))
                context = category[1] 
                matchedFolders.append(folderId)
            
    return [foldersWithTitles, matchedFolders]




